
#ifndef __L_SYSTEM_H
#define __L_SYSTEM_H


#include "../util/Pool.h"
#include "../util/Random.h"
#include "../util/ProgressUtil.h"
#include <vector>


/**
 *  A stochastic Lindenmayer system.
 *  @param SPIKE A concrete element (branch) class.
 *  @param RULE A rule class.
 *  @param VAR Number of element classes for rules derivation.
 */
template<class SPIKE, class RULE, int VAR>
class CLSystem {
public:
	enum { MAX_SIZE = 1000000 };
	typedef std::vector<RULE*>	TRuleVector;
public:
	CLSystem( CRandom& random, TRuleVector* rules );

	/** Evolutes the system one generation forward. */
	void evolve();

	/** Fills given vector with all spikes. */
	void getSpikes( std::vector<const SPIKE*>& container ) const;

	int getSize() const;

	/** For prepopulating with initial spikes. */
	void addSpike( int var, const SPIKE& spike );

private:
	int getPassiveSize() const;
	typedef CPool<SPIKE>		TSpikePool;
private:
	CRandom&	mRandom;
	TSpikePool	mSpikes1[VAR];
	TSpikePool	mSpikes2[VAR];
	bool		mActiveIs1;
	TRuleVector* mRules;
};



template<class SPIKE, class RULE, int VAR>
CLSystem<SPIKE,RULE,VAR>::CLSystem( CRandom& random, TRuleVector* rules )
:	mRandom( random ),
	mActiveIs1( true ),
	mRules( rules )
{
};

template<class SPIKE, class RULE, int VAR>
void CLSystem<SPIKE,RULE,VAR>::addSpike( int var, const SPIKE& spike )
{
	mSpikes1[var].add( spike );
}


template<class SPIKE, class RULE, int VAR>
void CLSystem<SPIKE,RULE,VAR>::evolve()
{
	TSpikePool* spkActive = (mActiveIs1)?(mSpikes1):(mSpikes2);
	TSpikePool* spkPassive = (mActiveIs1)?(mSpikes2):(mSpikes1);


	CProgressUtil::begin();

	for( int i = 0; i < VAR; i++ ) {
		TRuleVector& rules = mRules[i];
		int cnt = 0;
		
		// go thru all spikes
		TSpikePool::iterator spkit;
		for( spkit = spkActive[i].begin(); spkit != spkActive[i].end(); ++spkit ) {
			
			// HACK
			if( !cnt ) {
				int sz = getPassiveSize();
				if( sz >= MAX_SIZE ) {
					while( spkit != spkActive[i].end() ) {
						spkPassive[i].add( *spkit );
						++spkit;
					}
					break;
				}
				CProgressUtil::render();
			}
			cnt++;
			cnt &= 1023;
			
			// randomly select a rule for spike
			float p = mRandom.getFloat();
			float acc = 0.0f;
			TRuleVector::iterator rulit;
			for( rulit = rules.begin(); rulit != rules.end(); ++rulit ) {
				acc += (*rulit)->getProb();
				if( acc >= p )
					break;
			}
			// apply rule
			if( rulit != rules.end() )
				(*rulit)->apply( (*spkit), VAR, spkPassive );
		}
		
		// clear old ones
		spkActive[i].clear();
	}

	// switch old passives to new actives
	mActiveIs1 = !mActiveIs1;

	CProgressUtil::end();
};


template<class SPIKE, class RULE, int VAR>
void CLSystem<SPIKE,RULE,VAR>::getSpikes( std::vector<const SPIKE*>& container ) const
{
	const TSpikePool* spkActive = (mActiveIs1)?(mSpikes1):(mSpikes2);

	for( int i = 0; i < VAR; i++ ) {
		// go thru all spikes
		TSpikePool::const_iterator spkit;
		for( spkit = spkActive[i].begin(); spkit != spkActive[i].end(); ++spkit )
			container.push_back( &(*spkit) );
	}
}

template<class SPIKE, class RULE, int VAR>
int CLSystem<SPIKE,RULE,VAR>::getSize() const
{
	const TSpikePool* spkActive = (mActiveIs1)?(mSpikes1):(mSpikes2);
	
	int sz = 0;
	for( int i = 0; i < VAR; i++ )
		sz += spkActive[i].size();
	return sz;
};

template<class SPIKE, class RULE, int VAR>
int CLSystem<SPIKE,RULE,VAR>::getPassiveSize() const
{
	const TSpikePool* spkPassive = (mActiveIs1)?(mSpikes2):(mSpikes1);
	
	int sz = 0;
	for( int i = 0; i < VAR; i++ )
		sz += spkPassive[i].size();
	return sz;
};


#endif
