
#include "Drawing.h"

#include <cstring>
#include <cstdlib>

#define SWAP_INTS(a,b) { a ^= b; b ^= a; a ^= b; }


// ---------------------------------------------------------------------

CDrawing::CDrawing()
{
	// transparency LUT
	for( int i = 0; i < 256; i++ ) {
		for( int j = 0; j < TRANSPARENCY_LEVELS; j++ )
			mTransparencyLUT[i*TRANSPARENCY_LEVELS+j] = (unsigned char)( (float)i * j / (TRANSPARENCY_LEVELS-1) );
	}
	// steps (y is 0 by default)
	mXStep[0] = -(int)sizeof(CCanvas::pixel); mXStep[1] = 0; mXStep[2] = sizeof(CCanvas::pixel);
	mYStep[0] = mYStep[1] = mYStep[2] = 0;
}

CDrawing::~CDrawing()
{
}


void CDrawing::prepareCanvas( CCanvas& to )
{
	mYStep[0] = -to.getWidth() * sizeof( CCanvas::pixel );
	mYStep[2] =  to.getWidth() * sizeof( CCanvas::pixel );
}


// ---------------------------------------------------------------------
//  Blitting

void CDrawing::blit( CCanvas& to, const CCanvas& from, int x, int y )
{
	int bytes = from.getWidth() * 4;
	for( int i = 0; i < from.getHeight(); i++ )
		memcpy( &to.getRow(y+i)[x], from.getRow(i), bytes );
}


void CDrawing::blit( CCanvas& to, const CCanvas& from, int x, int y, int subx, int suby, int subw, int subh )
{
	int bytes = subw * 4;
	for( int i = 0; i < subh; i++ )
		memcpy( &to.getRow(y+i)[x], &from.getRow(suby+i)[subx], bytes );
}

void CDrawing::blitBlendAdd( CCanvas& to, const CCanvas& from, int x, int y, int subx, int suby, int subw, int subh )
{
	for( int i = 0; i < subh; i++ ) {
		const CCanvas::pixel* pfrom = from.getRow(suby+i) + subx;
		CCanvas::pixel* pto = to.getRow(y+i) + x;
		for( int j = 0; j < subw; j++ ) {
			CCanvas::pixel blend = CCanvas::rgbAdd( *pto, *pfrom );
			*pto = blend;
			pfrom++;
			pto++;
		}
	}
}

void CDrawing::blitBlendSub( CCanvas& to, const CCanvas& from, int x, int y, int subx, int suby, int subw, int subh )
{
	for( int i = 0; i < subh; i++ ) {
		const CCanvas::pixel* pfrom = from.getRow(suby+i) + subx;
		CCanvas::pixel* pto = to.getRow(y+i) + x;
		for( int j = 0; j < subw; j++ ) {
			CCanvas::pixel blend = CCanvas::rgbSub( *pto, *pfrom );
			*pto = blend;
			pfrom++;
			pto++;
		}
	}
}


// ---------------------------------------------------------------------
//  Lines
//  Fixed point DDA line drawing algorithm

void CDrawing::lineBlend( CCanvas& to, int x1, int y1, int x2, int y2, CCanvas::pixel color, int alpha ) const
{
	// premultiply alpha
	int invAlpha = TRANSPARENCY_LEVELS-1 - alpha;
	unsigned char cr = calcTransparency( (color&0xFF0000)>>16, alpha );
	unsigned char cg = calcTransparency( (color&0x00FF00)>>8 , alpha );
	unsigned char cb = calcTransparency( (color&0x0000FF)    , alpha );
	color = CCanvas::rgb( cr, cg, cb );
	
	// check zero len
	if( x1==x2 && y1==y2 ) {
		pixelBlendPremult( to, x1, y1, color, invAlpha );
		return;
	}

	int x, y;
	CCanvas::pixel* p;
	
	int dx = abs(x1-x2);
	int dy = abs(y1-y2);
	
	if( dx >= dy ) {
		
		//
		// more-horizontal line case

		int fy;
		int fdy = (dy << 16)/(dx);
		if( y2 < y1 )
			fdy = -fdy;

		if( x2 > x1 ) {
			fy = (y1 << 16) + (1<<15);
			p = to.getRow( y1 ) + x1;
			y = y1;
			for( x = x1; x <= x2; x++ ) {
				pixelBlendPremult( *p, color, invAlpha );
				fy += fdy;
				int yy = fy >> 16;
				((char*&)p) += mYStep[ yy-y+1 ];
				p++;
				y = yy;
			}
		} else {
			fy = (y2 << 16) + (1<<15);
			p = to.getRow( y2 ) + x2;
			y = y2;
			for( x = x2; x <= x1; x++ ) {
				pixelBlendPremult( *p, color, invAlpha );
				fy -= fdy;
				int yy = fy >> 16;
				((char*&)p) += mYStep[ yy-y+1 ];
				p++;
				y = yy;
			}
		}
	} else {

		//
		// more-vertical line case

		int fx;
		int fdx = (dx << 16)/(dy);
		if( x2 < x1 )
			fdx = -fdx;
		
		if( y2 > y1 ) {
			fx = (x1 << 16) + (1<<15);
			p = to.getRow( y1 ) + x1;
			x = x1;
			for( y = y1; y <= y2; y++ ) {
				pixelBlendPremult( *p, color, invAlpha );
				fx += fdx;
				int xx = fx >> 16;
				((char*&)p) += mXStep[ xx-x+1 ];
				((char*&)p) += mYStep[ 2 ];
				x = xx;
			}
		} else {
			fx = (x2 << 16) + (1<<15);
			p = to.getRow( y2 ) + x2;
			x = x2;
			for( y = y2; y <= y1; y++ ) {
				pixelBlendPremult( *p, color, invAlpha );
				fx -= fdx;
				int xx = fx >> 16;
				((char*&)p) += mXStep[ xx-x+1 ];
				((char*&)p) += mYStep[ 2 ];
				x = xx;
			}
		}
	}
}


void CDrawing::lineBlendZ( CCanvas& to, int x1, int y1, int x2, int y2, CCanvas::pixel color, int alpha, CCanvas& depth, float z1, float z2 ) const
{
	// premultiply alpha
	int invAlpha = TRANSPARENCY_LEVELS-1 - alpha;
	unsigned char cr = calcTransparency( (color&0xFF0000)>>16, alpha );
	unsigned char cg = calcTransparency( (color&0x00FF00)>>8 , alpha );
	unsigned char cb = calcTransparency( (color&0x0000FF)    , alpha );
	color = CCanvas::rgb( cr, cg, cb );
	
	// check zero len
	if( x1==x2 && y1==y2 ) {
		pixelBlendPremult( to, x1, y1, color, invAlpha );
		*( depth.getFPRow( y1 ) + x1 ) = z1;
		return;
	}

	int x, y;
	CCanvas::pixel* p;
	
	int dx = abs(x1-x2);
	int dy = abs(y1-y2);

	if( dx >= dy ) {
		
		//
		// more-horizontal line case

		int fy;
		int fdy = (dy << 16)/(dx);
		if( y2 < y1 )
			fdy = -fdy;

		float z;
		float dz;

		if( x2 > x1 ) {
			z = z1;
			dz = (z2-z1) / (float)dx;

			fy = (y1 << 16) + (1<<15);
			p = to.getRow( y1 ) + x1;
			y = y1;
			for( x = x1; x <= x2; x++ ) {
				pixelBlendPremult( *p, color, invAlpha );
				*(depth.getFPRow(y) + x) = z;
				fy += fdy;
				int yy = fy >> 16;
				z += dz;
				((char*&)p) += mYStep[ yy-y+1 ];
				p++;
				y = yy;
			}
		} else {
			z = z2;
			dz = (z1-z2) / (float)dx;

			fy = (y2 << 16) + (1<<15);
			p = to.getRow( y2 ) + x2;
			y = y2;
			for( x = x2; x <= x1; x++ ) {
				pixelBlendPremult( *p, color, invAlpha );
				*(depth.getFPRow(y) + x) = z;
				fy -= fdy;
				int yy = fy >> 16;
				z += dz;
				((char*&)p) += mYStep[ yy-y+1 ];
				p++;
				y = yy;
			}
		}
	} else {

		//
		// more-vertical line case

		int fx;
		int fdx = (dx << 16)/(dy);
		if( x2 < x1 )
			fdx = -fdx;
		
		float z;
		float dz;

		if( y2 > y1 ) {
			z = z1;
			dz = (z2-z1) / (float)dy;

			fx = (x1 << 16) + (1<<15);
			p = to.getRow( y1 ) + x1;
			x = x1;
			for( y = y1; y <= y2; y++ ) {
				pixelBlendPremult( *p, color, invAlpha );
				*(depth.getFPRow(y) + x) = z;
				fx += fdx;
				int xx = fx >> 16;
				z += dz;
				((char*&)p) += mXStep[ xx-x+1 ];
				((char*&)p) += mYStep[ 2 ];
				x = xx;
			}
		} else {
			z = z2;
			dz = (z1-z2) / (float)dy;

			fx = (x2 << 16) + (1<<15);
			p = to.getRow( y2 ) + x2;
			x = x2;
			for( y = y2; y <= y1; y++ ) {
				pixelBlendPremult( *p, color, invAlpha );
				*(depth.getFPRow(y) + x) = z;
				fx -= fdx;
				int xx = fx >> 16;
				z += dz;
				((char*&)p) += mXStep[ xx-x+1 ];
				((char*&)p) += mYStep[ 2 ];
				x = xx;
			}
		}
	}
}

