
#ifndef __DRAWING_H
#define __DRAWING_H

#include "Canvas.h"
#include "../util/MemUtil.h"

class CDrawing {
public:
	enum { TRANSPARENCY_LEVELS = 32 };
public:
	CDrawing();
	~CDrawing();

	void prepareCanvas( CCanvas& to );


	/** Blits "from" into "to", with top-left corner at x,y. NO checking! */
	static void blit( CCanvas& to, const CCanvas& from, int x, int y );

	/** Blits sub image of "from" into "to", with top-left corner at x,y. NO checking! */
	static void blit( CCanvas& to, const CCanvas& from, int x, int y, int subx, int suby, int subw, int subh );

	/** Blits sub image of "from" into "to", with top-left corner at x,y; add blend. NO checking! */
	static void blitBlendAdd( CCanvas& to, const CCanvas& from, int x, int y, int subx, int suby, int subw, int subh );

	/** Blits sub image of "from" into "to", with top-left corner at x,y; sub blend. NO checking! */
	static void blitBlendSub( CCanvas& to, const CCanvas& from, int x, int y, int subx, int suby, int subw, int subh );
	
	/** Alpha pixel. NO checking! */
	void pixelBlend( CCanvas& to, int x, int y, CCanvas::pixel color, int alpha ) const;

	/** Alpha pixel with remultiplier alpha color. NO checking! */
	void pixelBlendPremult( CCanvas& to, int x, int y, CCanvas::pixel color, int invAlpha ) const;
	
	/** Alpha line. call prepareCanvas() before! NO checking! */
	void lineBlend( CCanvas& to, int x1, int y1, int x2, int y2, CCanvas::pixel color, int alpha ) const;

	/** Alpha line with depth. call prepareCanvas() before! NO checking! */
	void lineBlendZ( CCanvas& to, int x1, int y1, int x2, int y2, CCanvas::pixel color, int alpha, CCanvas& depth, float z1, float z2 ) const;
	
	/** Fills canvas with single color. */
	static void fill( CCanvas& to, CCanvas::pixel color )
		{ memutil::fill32( to.getData(), color, to.getWidth()*to.getHeight() ); }
	/** Fills canvas with single float value. */
	static void fillFP( CCanvas& to, float value )
		{ memutil::fill32( to.getData(), *((unsigned int*)&value), to.getWidth()*to.getHeight() ); }

	void pixelBlend( CCanvas::pixel& dst, CCanvas::pixel color, int alpha ) const;
	void pixelBlendPremult( CCanvas::pixel& dst, CCanvas::pixel color, int invAlpha ) const;

private:
	unsigned char calcTransparency( unsigned char original, int level ) const
		{ return mTransparencyLUT[ ((int)original) * TRANSPARENCY_LEVELS + level ]; }
	
private:
	unsigned char	mTransparencyLUT[256*TRANSPARENCY_LEVELS];
	int		mXStep[3];
	int		mYStep[3];
};



inline void CDrawing::pixelBlend( CCanvas& to, int x, int y, CCanvas::pixel color, int alpha ) const
{
	CCanvas::pixel& dst = *( to.getRow( y ) + x );
	int invAlpha = TRANSPARENCY_LEVELS-1 - alpha;

	unsigned char r = calcTransparency( (color&0xFF0000)>>16, alpha ) + calcTransparency( (dst&0xFF0000)>>16, invAlpha );
	unsigned char g = calcTransparency( (color&0x00FF00)>>8 , alpha ) + calcTransparency( (dst&0x00FF00)>>8 , invAlpha );
	unsigned char b = calcTransparency( (color&0x0000FF)    , alpha ) + calcTransparency( (dst&0x0000FF)    , invAlpha );

	dst = CCanvas::rgb( r, g, b );
};

inline void CDrawing::pixelBlend( CCanvas::pixel& dst, CCanvas::pixel color, int alpha ) const
{
	int invAlpha = TRANSPARENCY_LEVELS-1 - alpha;

	unsigned char r = calcTransparency( (color&0xFF0000)>>16, alpha ) + calcTransparency( (dst&0xFF0000)>>16, invAlpha );
	unsigned char g = calcTransparency( (color&0x00FF00)>>8 , alpha ) + calcTransparency( (dst&0x00FF00)>>8 , invAlpha );
	unsigned char b = calcTransparency( (color&0x0000FF)    , alpha ) + calcTransparency( (dst&0x0000FF)    , invAlpha );

	dst = CCanvas::rgb( r, g, b );
};

inline void CDrawing::pixelBlendPremult( CCanvas& to, int x, int y, CCanvas::pixel color, int invAlpha ) const
{
	CCanvas::pixel& dst = *( to.getRow( y ) + x );

	unsigned char r = ((color&0xFF0000)>>16) + calcTransparency( (dst&0xFF0000)>>16, invAlpha );
	unsigned char g = ((color&0x00FF00)>>8 ) + calcTransparency( (dst&0x00FF00)>>8 , invAlpha );
	unsigned char b = ((color&0x0000FF)    ) + calcTransparency( (dst&0x0000FF)    , invAlpha );
	
	dst = CCanvas::rgb( r, g, b );
};

inline void CDrawing::pixelBlendPremult( CCanvas::pixel& dst, CCanvas::pixel color, int invAlpha ) const
{
	unsigned char r = ((color&0xFF0000)>>16) + calcTransparency( (dst&0xFF0000)>>16, invAlpha );
	unsigned char g = ((color&0x00FF00)>>8 ) + calcTransparency( (dst&0x00FF00)>>8 , invAlpha );
	unsigned char b = ((color&0x0000FF)    ) + calcTransparency( (dst&0x0000FF)    , invAlpha );
	
	dst = CCanvas::rgb( r, g, b );
};


#endif
